{% extends "!layout.html" %}

{%- block extrahead %}
 <script type="text/javascript">
       MathJax.Hub.Config({
           displayAlign: "left",
           displayIndent: "2.5em"
       });
  </script>
{% endblock %}