@import url("haiku.css");

div.contents { 
    float: right; 
    background: Cornsilk;
    margin: 0.5em 0.5em;
}

div.contents p.topic-title {
    display: none;
}

div.contents ul {
    padding: 0em;
}

div.contents ul > li {
    background: none;
    padding: 0 0 0 1em;
}
